/* C function to read a cad4 reflection and place it into a 
 * struct cad4_reflection.  The function take a FILE pointer
 * to the CAD4 data file and a pointer to an int in which the
 * total number of reflections processed are stored.
 * Written by:
 *	Paul D. Boyle
 *	Box 8204
 *	Department of Chemistry
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *	boyle@laue.chem.ncsu.edu
 */
#define _POSIX_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* local headers */

#include "fildat.h"

struct cad4_reflection *GetCAD4Reflection( FILE *cad4_file,
                                           int *n_ref, int *i_dumps ) 
{
    struct cad4_reflection *refl = NULL,
                           *tmp_ref = NULL;

    char                   read_in[CAD4_LINE_LEN],
                           junk[CAD4_LINE_LEN],
                           *ptr;

    int     index = 0,  /* current reflection index */
            record = 0,
#ifdef USING_PROFILES
            n_dumps = N_DUMPS,
#endif
            j_junk,
            i_junk;   /* disregard subsequent reads of Nrefl */

#ifdef USING_PROFILES
    if( NULL != i_dumps ) n_dumps = *i_dumps;
#endif

    while( NULL != fgets(read_in, CAD4_LINE_LEN, cad4_file) ) {
        if( NULL != (ptr = strchr(read_in, '\n')) )
            *ptr = '\0';
        if( NULL != (ptr = strchr(read_in, '\r')) )
            *ptr = '\0';
        if( 0 == strncmp(read_in, END_DATA_FILE, strlen(END_DATA_FILE)) )
            break;
/* check to see that we are getting data file records rather than the
 * listing of the fcryst file reflection list at the beginning of the    
 * data file.
 */
        strncpy( junk, read_in, 4 );
        junk[4] = '\0';
        if( NULL == strpbrk( junk, "0123456789" ) ) continue;


        sscanf( read_in, "%d %s", &record, junk );
        if(  1 == record  ) {
            (*n_ref)++;
            index = *n_ref - 1;
            tmp_ref = realloc( refl, *n_ref * sizeof(struct cad4_reflection) );
            if( !tmp_ref ) {
              fprintf( stderr, "Memory Allocation Error: %s:%d\n", 
                       __FILE__, __LINE__ );
              fprintf( stderr,
                     "Warning: only %d CAD4 reflections successfully read in.\n", 
                     *n_ref - 1 );
              return refl;
            }
            refl = tmp_ref;
#ifdef USING_PROFILES
            refl[index].profile = NULL;
#endif
        }
        switch( record ) {
          case  1:
             sscanf( read_in, "%d %d %f %f %f %6s %f %f %f %f %f", &i_junk,
                     &refl[index].nrefl, &refl[index].hkl[H], &refl[index].hkl[K],
                     &refl[index].hkl[L], refl[index].code, &refl[index].psi,
                     &refl[index].npi, &refl[index].bkg_left, 
                     &refl[index].peak, &refl[index].bkg_right );  
             break;
          case  2:
             sscanf( read_in, "%d %d %f %f %f %f %f %u %d", &i_junk, &j_junk,
                     &refl[index].theta, &refl[index].phi_k,
                     &refl[index].omega_k, &refl[index].kappa, 
                     &refl[index].width, &refl[index].xrayt,
                     &refl[index].fridl );
             break;
#ifdef USING_PROFILES
          case  3:
             refl[index].profile = malloc( N_DUMPS * sizeof(float) );
             if( !refl[index].profile ) {
                 fprintf( stderr, "Memory Allocation Error: %s:%d\n", 
                          __FILE__, __LINE__ );
                 fprintf( stderr, "Cannot Allocate memory for profile for"
                          " reflection %d\n", *n_ref );
                 break;
             }
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[0],  &refl[index].profile[1],
                     &refl[index].profile[2],  &refl[index].profile[3],
                     &refl[index].profile[4],  &refl[index].profile[5],
                     &refl[index].profile[6],  &refl[index].profile[7],
                     &refl[index].profile[8],  &refl[index].profile[9],
                     &refl[index].profile[10], &refl[index].profile[11] );
             break;
          case  4:
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[12], &refl[index].profile[13],
                     &refl[index].profile[14], &refl[index].profile[15],
                     &refl[index].profile[16], &refl[index].profile[17],
                     &refl[index].profile[18], &refl[index].profile[19],
                     &refl[index].profile[20], &refl[index].profile[21],
                     &refl[index].profile[22], &refl[index].profile[23] );
             break;
          case  5:
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[24], &refl[index].profile[25],
                     &refl[index].profile[26], &refl[index].profile[27],
                     &refl[index].profile[28], &refl[index].profile[29],
                     &refl[index].profile[30], &refl[index].profile[31],
                     &refl[index].profile[32], &refl[index].profile[33],
                     &refl[index].profile[34], &refl[index].profile[35] );
             break;
          case  6:
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[36], &refl[index].profile[37],
                     &refl[index].profile[38], &refl[index].profile[39],
                     &refl[index].profile[40], &refl[index].profile[41],
                     &refl[index].profile[42], &refl[index].profile[43],
                     &refl[index].profile[44], &refl[index].profile[45],
                     &refl[index].profile[46], &refl[index].profile[47] );
             break;
          case  7:
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[48], &refl[index].profile[49],
                     &refl[index].profile[50], &refl[index].profile[51],
                     &refl[index].profile[52], &refl[index].profile[53],
                     &refl[index].profile[54], &refl[index].profile[55],
                     &refl[index].profile[56], &refl[index].profile[57],
                     &refl[index].profile[58], &refl[index].profile[59] );
             break;
          case  8:
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[60], &refl[index].profile[61],
                     &refl[index].profile[62], &refl[index].profile[63],
                     &refl[index].profile[64], &refl[index].profile[65],
                     &refl[index].profile[66], &refl[index].profile[67],
                     &refl[index].profile[68], &refl[index].profile[69],
                     &refl[index].profile[70], &refl[index].profile[71] );
             break;
          case  9:
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[72], &refl[index].profile[73],
                     &refl[index].profile[74], &refl[index].profile[75],
                     &refl[index].profile[76], &refl[index].profile[77],
                     &refl[index].profile[78], &refl[index].profile[79],
                     &refl[index].profile[80], &refl[index].profile[81],
                     &refl[index].profile[82], &refl[index].profile[83] );
             break;
          case 10:
             sscanf( read_in, "%d %d %4f%4f%4f %4f%4f%4f %4f%4f%4f %4f%4f%4f", 
                     &i_junk, &j_junk,
                     &refl[index].profile[84], &refl[index].profile[85],
                     &refl[index].profile[86], &refl[index].profile[87],
                     &refl[index].profile[88], &refl[index].profile[89],
                     &refl[index].profile[90], &refl[index].profile[91],
                     &refl[index].profile[92], &refl[index].profile[93],
                     &refl[index].profile[94], &refl[index].profile[95] );
             break;
#endif
          default:
             break;
        }

    }
    rewind( cad4_file );
    return refl;
}
