/* program to read a CAD4 data file, and output theta value and 
 * net intensity.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	August 1999
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */
#define _XOPEN_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <assert.h>
#include <getopt.h>

#include "fildat.h" 

/* prototypes */
void usage( char *name );
int theta_sort( const void *, const void * );
struct cad4_reflection *get_reflection_data( FILE *cad4_file,
                                             int *n_refs,
                                             float *old_lambda,
                                             float *new_lambda );
float net_intensity( float att, struct cad4_reflection *r, float *sigma );
float calc_two_theta( float old_lambda, float new_lambda, float theta );
int evaluate( float w1, float w2, float angle );

struct theta_net_int {
       float hkl[3];
       float two_theta;
       float i_net;
       float sigma_inet;
       };

#define DEFAULT_OUT_FILE "CAD2PWD.OUT"

#define DEFAULT_OUTPUT_FMT  "%8.3f %9.2f\n"
#define VERBOSE_OUTPUT_FMT  "%8.3f %9.2f %4.0f%4.0f%4.0f\n"
#define WGHT  0.6666667
#define H 0
#define K 1
#define L 2

#ifndef PI
#define PI 3.14159265
#endif

#define DEG2RAD(deg) ((PI / 180.0) * (deg))
#define RAD2DEG(rad) ((180.0 / PI) * (rad))

int main( int argc, char **argv )
{
    FILE *data_file,
         *out = stdout;
    struct cad4_experimental exp;
    struct cad4_reflection   *refl = NULL;
    struct theta_net_int     *data = NULL;
    int                      i,
                             n_ref = 0,
                             verbose = 0,
                             no_use_cad4_lambda = 0,
                             i_opt;
    size_t                   len;

    char *output_file_name = NULL;
    char *output_format = DEFAULT_OUTPUT_FMT;
    float lambda_orig,
          lambda_new;

/* get the command line options parsed */

    if( 1 == argc ) {
        usage( argv[0] );
        exit( EXIT_FAILURE );
    }
    while( (i_opt = getopt(argc, argv, "vl:o:")) != EOF ) {
      switch( i_opt ) {
        case 'v':
           verbose = 1;
           output_format = VERBOSE_OUTPUT_FMT;
           break;
        case 'l':
           errno = 0;
           lambda_new = (float)strtod( optarg, (char **)NULL );
           if( errno != 0 ) {
              fputs( "Lambda value is no good!\n", stderr );
              perror( optarg );
              exit( EXIT_FAILURE );
           }
           no_use_cad4_lambda = 1;
           break;
        case 'o':
           len = strlen( optarg );
           output_file_name = malloc( len + 1 );
           if( !output_file_name ) {
                fprintf( stderr, "malloc() failed -- %s:%d. Writing output to stdout.\n", 
                                 __FILE__, __LINE__ );
                
                break;
           }
           else {
               strcpy( output_file_name, optarg );
               out = fopen( output_file_name, "w" );
               if( !out ) {
                   perror( output_file_name );
                   exit( EXIT_FAILURE );
               }
           }

           break;
      }
    }

/* open the cad4 data file */
    data_file = fopen( argv[argc-1], "r" );
    if( !data_file ) {
        perror( argv[argc-1] );
        usage( argv[0] );
        exit( EXIT_FAILURE );
    }

/* get the experiemental section at the front of the file */
    GetCAD4Experimental( data_file, &exp );
    lambda_orig = WGHT * exp.lambda1 + (1.0 - WGHT) * exp.lambda2;

/* refl contains the data set with the intensity checks stripped out,
 * n_ref contains the number of reflections in the dataset.
 */
    if( no_use_cad4_lambda )
        refl = get_reflection_data( data_file, &n_ref, &lambda_orig,
                                                       &lambda_new );
    else
        refl = get_reflection_data( data_file, &n_ref, NULL,
                                                       NULL );
        

    fclose( data_file );

    data = malloc( n_ref * sizeof( *data ) );
    if( !data ) {
        fprintf( stderr, "malloc() failed -- %s:%d. Exiting.\n", 
                          __FILE__, __LINE__ );
        exit( EXIT_FAILURE );
    }

    for( i = 0; i < n_ref; i++ ) {
         data[i].sigma_inet = 0.0;
         memcpy( data[i].hkl, refl[i].hkl, 3 * sizeof(float) );
         if( 1 == no_use_cad4_lambda )
             data[i].two_theta = calc_two_theta( lambda_orig,
                                          lambda_new, refl[i].theta );
         else
             data[i].two_theta = 2.0 * refl[i].theta;

         data[i].i_net = net_intensity(  exp.attenuator, &refl[i], NULL );
    }

    qsort( data, n_ref, sizeof(data[0]), theta_sort );

    for( i = 0; i < n_ref; i++ ) {
         if( verbose )
             fprintf( out, output_format, data[i].two_theta,
                           data[i].i_net, data[i].hkl[H],
                           data[i].hkl[K], data[i].hkl[L] );
         else
             fprintf( out, output_format, data[i].two_theta, data[i].i_net );
    }

    free( refl );
    free( data );
    fclose( out );
    exit( EXIT_SUCCESS );
}   

/* get_reflection_data() -- function which retrieves the raw cad4 data
 * sans the intensity checks.
 */

struct cad4_reflection *get_reflection_data( FILE *cad4_file, int *n_refs,
                                             float *old_lambda, float *new_lambda)
{
    struct cad4_reflection *raw  = NULL,
                           *data = NULL; 

    int n_reflections = 0,
        nr            = 0, 
        limit_flag    = 1,
        i;

    *n_refs = 0;

    raw = GetCAD4Reflection( cad4_file, &n_reflections, NULL );

    for( i = 0; i < n_reflections; i++ ) {
         struct cad4_reflection *tmp;
         if( raw[i].code[0] == 'I' ) continue;
         if( NULL != old_lambda && NULL != new_lambda )
             limit_flag = evaluate( *old_lambda, *new_lambda, raw[i].theta );

         if( -1 == limit_flag ) continue;

         nr++;
         tmp = realloc( data, nr * sizeof(*data) );
         if( !tmp ) {
             fprintf( stderr, "realloc() -- failure: %s:%d\n", 
                              __FILE__, __LINE__ );
             fprintf( stderr,
                      "**warning** only %d CAD4 reflections processed.\n",
                      nr - 1);

             *n_refs = nr - 1;
             return data;
         }
         data = tmp; 
         data[nr-1] = raw[i];
    }      
    free( raw );
    *n_refs = nr;
    return data;
}

/* theta_sort() -- function to be passed to qsort for sorting 
 * the reflections.
 */

int theta_sort( const void *a, const void *b )
{
    struct theta_net_int   aa,
                           bb;

    aa = *(struct theta_net_int *)a;
    bb = *(struct theta_net_int *)b;

    if( aa.two_theta < bb.two_theta )
        return -1;
    else if( aa.two_theta > bb.two_theta )
        return 1;

   return 0;
}

void usage( char *name )
{
    int i = 0;
    static char *options[] = {
           "-v   verbose output, prints hkl's",
           "-l<wavelength>  specify wavelength to be used",
           "-o<output_file_name>  specify output file.  Default output to console",
           "" };
    fprintf( stderr, "Usage:  %s [options] <cad4_data_file_name>\n", name );
    while( *options[i] ) {
        fprintf(stderr, "%s\n", options[i] );
        i++;
    }

    return;
}

     
/* This function calculates net intensities and sigmas from
 * a CAD4 .dat file. 
 */


float net_intensity( float att, struct cad4_reflection *r, float *sigma )
{
    float sum, 
          i_net = 0.0;
    float atten = 1.0;

    if( r->npi < 0 ) {
        r->npi = -r->npi;
        atten = att;
    }
    sum = atten * ( r->bkg_left + r->bkg_right );
    r->peak = atten * r->peak;
    
    i_net = ( r->peak - 2.0 * sum ) / r->npi;
    if( NULL != sigma )
       *sigma =  (float)sqrt((double)(r->peak + 4.0 * sum))  / r->npi;

    return i_net; 
}

float calc_two_theta( float old_lambda, float new_lambda, float theta )
{
   double sin_theta2;

   float   two_theta;

   sin_theta2 = new_lambda / old_lambda * sin( DEG2RAD(theta) );
   two_theta = 2.0 * RAD2DEG( asin(sin_theta2) );

   return two_theta;
}
/* evaluate() -- this function evaluates whether or not a reflection observed
 * at 'angle' using wavelength 'w1' will be observable if wavelength 'w2' is
 * is used.  If the reflection would be observable then the function 
 * returns 0, else it returns -1
 */
int evaluate( float w1, float w2, float angle )
{
    float limit;
  
    limit = w2 / w1 * sin( DEG2RAD(angle) );

    return 1.0 > limit ? 0 : -1;
}

   

